using Microsoft.ML;
using Microsoft.ML.Data;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

using TestLibrary;

namespace TestProject
{
    /// <summary>
    ///  
    /// </summary>
    public static class ModelBuilder
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// Ʒ   
        /// </summary>
        private static string TRAINING_DATA_FILE_PATH = @"DATA\taxi-fare-train.csv";

        /// <summary>
        /// ġ   
        /// </summary>
        private static string WEIGHT_MODEL_FILE_PATH = @"WEIGHT\MLModel.zip";

        /// <summary>
        /// ML ؽƮ
        /// </summary>
        private static MLContext _mlContext = new MLContext(seed : 1);

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region  ϱ - CreateModel()

        /// <summary>
        ///  ϱ
        /// </summary>
        public static void CreateModel()
        {
            IDataView trainingDataView = _mlContext.Data.LoadFromTextFile<ModelInput>
            (
                path          : TRAINING_DATA_FILE_PATH,
                hasHeader     : true,
                separatorChar : ',',
                allowQuoting  : true,
                allowSparse   : false
            );

            IEstimator<ITransformer> trainingPipeline = BuildTrainingPipeline(_mlContext);

            ITransformer mlModel = TrainModel(_mlContext, trainingDataView, trainingPipeline);

            Evaluate(_mlContext, trainingDataView, trainingPipeline);

            SaveModel(_mlContext, mlModel, WEIGHT_MODEL_FILE_PATH, trainingDataView.Schema);
        }

        #endregion
        #region Ʒ   - BuildTrainingPipeline(mlContext)

        /// <summary>
        /// Ʒ  
        /// </summary>
        /// <param name="mlContext">ML ؽƮ</param>
        /// <returns>Ʒ </returns>
        public static IEstimator<ITransformer> BuildTrainingPipeline(MLContext mlContext)
        {
            var dataProcessPipeline = mlContext.Transforms.Categorical.OneHotEncoding
            (
                new[]
                {
                    new InputOutputColumnPair("vendor_id"   , "vendor_id"   ),
                    new InputOutputColumnPair("payment_type", "payment_type")
                }
            )
            .Append
            (
                mlContext.Transforms.Concatenate
                (
                    "Features",
                    new[] { "vendor_id", "payment_type", "rate_code", "passenger_count", "trip_distance" }
                )
            );

            var trainer = mlContext.Regression.Trainers.LightGbm(labelColumnName : "fare_amount", featureColumnName : "Features");

            var trainingPipeline = dataProcessPipeline.Append(trainer);

            return trainingPipeline;
        }

        #endregion
        #region  Ʒϱ - TrainModel(mlContext, trainingDataView, trainingPipeline)

        /// <summary>
        ///  Ʒϱ
        /// </summary>
        /// <param name="mlContext">ML ؽƮ</param>
        /// <param name="trainingDataView">Ʒ  </param>
        /// <param name="trainingPipeline">Ʒ </param>
        /// <returns></returns>
        public static ITransformer TrainModel(MLContext mlContext, IDataView trainingDataView, IEstimator<ITransformer> trainingPipeline)
        {
            Console.WriteLine(" Ʒ մϴ.");

            ITransformer model = trainingPipeline.Fit(trainingDataView);

            Console.WriteLine("Ʒ μ մϴ.");

            return model;
        }

        #endregion

        #region ȸ Ʈ ϱ - PrintRegressionMetrics(metrics)

        /// <summary>
        /// ȸ Ʈ ϱ
        /// </summary>
        /// <param name="metrics">Ʈ</param>
        public static void PrintRegressionMetrics(RegressionMetrics metrics)
        {
            Console.WriteLine($"ȸ  Ʈ");
            Console.WriteLine($"    LossFn        : {metrics.LossFunction        :0.##}");
            Console.WriteLine($"    R2 Score      : {metrics.RSquared            :0.##}");
            Console.WriteLine($"    Absolute loss : {metrics.MeanAbsoluteError   :#.##}");
            Console.WriteLine($"    Squared loss  : {metrics.MeanSquaredError    :#.##}");
            Console.WriteLine($"    RMS loss      : {metrics.RootMeanSquaredError:#.##}");
        }

        #endregion
        #region ȸ   Ʈ ϱ - PrintRegressionFoldsAverageMetrics(crossValidationResults)

        /// <summary>
        /// ȸ   Ʈ ϱ
        /// </summary>
        /// <param name="crossValidationResults">  </param>
        public static void PrintRegressionFoldsAverageMetrics(IEnumerable<TrainCatalogBase.CrossValidationResult<RegressionMetrics>> crossValidationResults)
        {
            IEnumerable<double> l1LossEnumerable       = crossValidationResults.Select(r => r.Metrics.MeanAbsoluteError   );
            IEnumerable<double> l2LossEnumerable       = crossValidationResults.Select(r => r.Metrics.MeanSquaredError    );
            IEnumerable<double> rmsEnumerable          = crossValidationResults.Select(r => r.Metrics.RootMeanSquaredError);
            IEnumerable<double> lossFunctionEnumerable = crossValidationResults.Select(r => r.Metrics.LossFunction        );
            IEnumerable<double> rSquaredEnumerable     = crossValidationResults.Select(r => r.Metrics.RSquared            );

            Console.WriteLine($"ȸ  Ʈ");
            Console.WriteLine($"  L1 Loss       : {l1LossEnumerable.Average()      :0.###}");
            Console.WriteLine($"  L2 Loss       : {l2LossEnumerable.Average()      :0.###}");
            Console.WriteLine($"  RMS           : {rmsEnumerable.Average()         :0.###}");
            Console.WriteLine($"  Loss Function : {lossFunctionEnumerable.Average():0.###}");
            Console.WriteLine($"  R-squared     : {rSquaredEnumerable.Average()    :0.###}");
        }

        #endregion
        #region   ϱ - GetAbsolutePath(relativePath)

        /// <summary>
        ///   ϱ
        /// </summary>
        /// <param name="relativePath"> </param>
        /// <returns> </returns>
        public static string GetAbsolutePath(string relativePath)
        {
            FileInfo fileInfo = new FileInfo(typeof(Program).Assembly.Location);

            string assemblyFolderPath = fileInfo.Directory.FullName;

            string fullPath = Path.Combine(assemblyFolderPath, relativePath);

            return fullPath;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region ϱ - Evaluate(mlContext, trainingDataView, trainingPipeline)

        /// <summary>
        /// ϱ
        /// </summary>
        /// <param name="mlContext">ML ؽƮ</param>
        /// <param name="trainingDataView">Ʒ  </param>
        /// <param name="trainingPipeline">Ʒ </param>
        private static void Evaluate(MLContext mlContext, IDataView trainingDataView, IEstimator<ITransformer> trainingPipeline)
        {
            Console.WriteLine(" Ȯ Ʈ ϱ   ");

            var crossValidationResults = mlContext.Regression.CrossValidate
            (
                trainingDataView,
                trainingPipeline,
                numberOfFolds   : 5,
                labelColumnName : "fare_amount"
            );

            PrintRegressionFoldsAverageMetrics(crossValidationResults);
        }

        #endregion
        #region  ϱ - SaveModel(mlContext, mlModel, modelRelativePath, modelInputSchema)

        /// <summary>
        ///  ϱ
        /// </summary>
        /// <param name="mlContext">ML ؽƮ</param>
        /// <param name="mlModel">ML </param>
        /// <param name="modelRelativePath">  </param>
        /// <param name="modelInputSchema"> Է Ű</param>
        private static void SaveModel(MLContext mlContext, ITransformer mlModel, string modelRelativePath, DataViewSchema modelInputSchema)
        {
            Console.WriteLine(" ");

            mlContext.Model.Save(mlModel, modelInputSchema, GetAbsolutePath(modelRelativePath));

            Console.WriteLine($" ߽ϴ : {GetAbsolutePath(modelRelativePath)}");
        }

        #endregion
    }
}